#------------------------------------------------------------------------------
#    59Tracker, weblog software for personal publisher.
#    Copyright (C) 2004-2010 Kaga, Hiroaki
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#------------------------------------------------------------------------------

package Lib::Keyword;

use strict;
use warnings;

use constant {
    LOCK_SH => 1,
    LOCK_EX => 2,
    LOCK_NB => 4,
    LOCK_UN => 8,
};

use URI::Escape;

use Lib::Logger;
use Lib::String;
use Lib::Language;

# コンストラクタ
sub new {
    my $self = {};

    use Lib::Conf;
    my $conf = Lib::Conf->new();

    $self->{system_dir} = $conf->get_system_dir();
    $self->{data_dir} = $conf->get_data_dir();
    $self->{docroot_dir} = $conf->get_docroot_dir();

    bless($self);
    return $self;
}

# キーワードの入力チェック（全角文字と半角英数のみキーワードとして使用可能）
sub check {
    my $self = shift;
    my $str = shift;

    # UTF-8からシフトJISに変換
    my $string = Lib::String->new();
    $str = $string->convert($str, 'utf8', 'sjis');

    my $len = length($str);
    for (my $i = 0; $i < $len; $i++) {
        my $ch = substr($str, $i, 1);
        if ($ch =~ /[\x81-\x9F\xE0-\xEF]/) { # 全角文字
            $i++;
            next;
        }
        elsif ($ch =~ /[\x20\x30-\x39\x41-\x5a\x61-\x7a]/) { # 半角英数
            next;
        }
        else {
            return 1;
        }
    }

    return 0;
}

# キーワードを追加する
sub add {
    my $self = shift;
    my ($item, $id) = @_;
    my $result = 0;
    my $notfound = 1;
    my @newdata = ();
    my $list;
    my $num;

    # キーワードデータの読み込み
    my $keywordfile = $self->{data_dir} . "/keyword/keyword.data";
    if (-f $keywordfile) {
        open my $infh, '<', $keywordfile;
        while (my $data = <$infh>) {
            chomp $data;
            if ($data ne '') {
                my ($keywordid, $keyworditem, $topiclist, $topicnum, $refernum) = split(/\,/, $data);
                if (uc($keyworditem) eq uc($item)) {
                    $list = $topiclist;
                    $num = $topicnum;

                    # 既に登録済かをチェック
                    my @idlist = split(/:/, $list);
                    my $hit = 0;
                    for my $i (0 .. $#idlist) {
                        if ($idlist[$i] == $id) {
                            $hit = 1;
                            last;
                        }
                    }
                    if (!$hit) {
                        if ($num > 0) {
                            $list .= ':';
                        }
                        $list .= $id;
                        $num++;

                        $topiclist = $list;
                        $topicnum = $num;
                    }

                    $data = "$keywordid,$keyworditem,$topiclist,$topicnum,$refernum";
                    $result = 1;
                    $notfound = 0;
                }
                push @newdata, $data;
           }
        }
        close $infh;
    }

    if ($notfound) {

        my ($topiclist, $topicnum) = ('', 0);
        $topiclist = $id;
        $topicnum = 1;

        # 新規キーワードIDを取得
        my $keywordid = 0;
        my $counterfile = $self->{system_dir} . "/keywordcounter.txt";
        open my $infh, '<', $counterfile;
        my $data = <$infh>;
        chomp $data;
        if ($data ne '') {
            $keywordid = $data;
        }
        close $infh;
        $keywordid++;

        # キーワードカウンターを更新
        open my $outfh, '>', $counterfile;
        flock $outfh, LOCK_EX;
        print {$outfh} "$keywordid\n";
        flock $outfh, LOCK_UN;
        close $outfh;

        $data = "$keywordid,$item,$topiclist,$topicnum,0";
        push @newdata, $data;
        $result = 1;
    }

    # キーワードデータファイルを更新
    open my $keywordfh, '>', $keywordfile;
    flock $keywordfh, LOCK_EX;
    foreach my $data (@newdata) {
        print {$keywordfh} "$data\n";
    }
    flock $keywordfh, LOCK_UN;
    close $keywordfh;

    # ファイルのパーミッションを変更
    chmod 0766, $keywordfile;

    return $result;
}

# キーワードをキーに指定されたIDを削除する
sub delete {
    my $self = shift;
    my ($item, $id) = @_;
    my $result = 0;
    my $notfound = 1;
    my @newdata = ();

    # キーワードデータの読み込み
    my $keywordfile = $self->{data_dir} . "/keyword/keyword.data";
    open my $infh, '<', $keywordfile;
    while (my $data = <$infh>) {
        chomp $data;
        if ($data ne '') {
            my ($keywordid, $keyworditem, $topiclist, $topicnum, $refernum) = split(/\,/, $data);
            if ($keyworditem eq $item) {

                my $list;
                my $num;
                $list = $topiclist;
                $num = $topicnum;

                # 登録されているかチェック
                my $newlist = '';
                my @idlist = split(/:/, $list);
                for my $i (0 .. $#idlist) {
                    if ($idlist[$i] != $id) {
                        if ($newlist ne '') {
                            $newlist .= ':';
                        }
                        $newlist .= $idlist[$i];
                    }
                    else {
                        $notfound = 0;
                    }
                }

                # 削除した場合は再編成
                if (!$notfound) {
                    $num = scalar(@idlist) - 1;
                    $topiclist = $newlist;
                    $topicnum = $num;
                    $data = "$keywordid,$item,$topiclist,$topicnum,$refernum";
                    $result = 1;
                }
            }
            push @newdata, $data;
        }
    }
    close $infh;

    if (!$notfound) {
        # キーワードデータファイルを更新
        open my $outfh ,'>', $keywordfile;
        flock $outfh, LOCK_EX;
        foreach my $data (@newdata) {
            print {$outfh} "$data\n";
        }
        flock $outfh, LOCK_UN;
        close $outfh;
    }

    return $result;
}

# キーワード項目配列からキーワードリストを作成する
sub list {
    my $self = shift;
    my @items = @_;

    my $itemlist = '';
    foreach my $item (@items) {
        my $keyworditem = $item;
        $item = uri_escape($item);
        $itemlist .= "<a href=\"$self->{docroot_dir}/?q=$item\" title=\"$keyworditem\">$keyworditem</a>&nbsp;";
    }

    my $result = '';
    if ($itemlist ne '') {
        my $language = Lib::Language->new();
        my $keywordlabel = $language->get_statictext('keyword');
        $result = "<b>$keywordlabel</b><br />$itemlist";
    }

    return $result;
}

# キーワードをリンクする
sub link_keyword {
    my $self = shift;
    my $keywords = shift;
    my $text = shift;

    if ($keywords) {
        my @keywordlist = split(/ /, $keywords);
        foreach my $keyword (@keywordlist) {
            my $enckeyword = uri_escape($keyword);
            my $link = "<a href=\"./?q=$enckeyword\">$keyword</a>";
            $text =~ s/$keyword(?!.*?<\/a>)/$link/gi;
        }
    }

    return $text;
}

# クラウドの取得
sub clouds {
    my $self = shift;
    my $format = shift; # 0:強調なし 1:強調あり
    my $itemlist = '';
    my $max = 0;    # 最大値
    my $p1;         # 平均値 + ((最大値 - 平均値) / 2)
    my $avrg;       # 平均値
    my $m1;         # 平均値 - ((平均値 - 最小値) / 2)
    my $min = 0;    # 最小値
    my $num;
    my @dataarray = ();
    my $itemnum = 0;
    my $totalnum = 0;

    my $keywordfile = $self->{data_dir} . "/keyword/keyword.data";
    if (!(-f $keywordfile)) {
        return $itemlist;
    }

    open my $keywordfh, '<', $keywordfile;
    while (my $data = <$keywordfh>) {
        chomp $data;
        if ($data ne '') {
            my ($keywordid, $keyworditem, $topiclist, $topicnum, $refernum) = (split(/\,/, $data))[0..4];
            $num = $topicnum;
            $num += $refernum;
            if ($num > $max) {
                $max = $num;
            }
            if ($num < $min) {
                $min = $num;
            }
            if ($num > 0) {
                $totalnum += $num;
                $itemnum++;
                push @dataarray, $data;
            }
        }
    }
    close $keywordfh;

    if ($itemnum > 0) {
        $avrg = $totalnum / $itemnum;
    }
    else {
        $avrg = 0;
    }
    $p1 = $avrg + (($max - $avrg) / 2);
    $m1 = $avrg - (($avrg - $min) / 2);

    foreach my $data (@dataarray) {
        my ($keywordid, $keyworditem, $topiclist, $topicnum, $refernum) = (split(/\,/, $data))[0..4];
        $num = $topicnum;
        $num += $refernum;

        if ($num > $p1) {
            my $item = $keyworditem;
            $item = uri_escape($item);
            $itemlist .= <<"END_ITEM";
<a href="./?q=$item" title="$keyworditem">$keyworditem</a>&nbsp;
END_ITEM
        }
    }

    if ($itemlist ne '') {
        my $language = Lib::Language->new();
        my $hotkeyword = $language->get_statictext('hot-keyword');
        $itemlist = "<h3>$hotkeyword</h3>$itemlist<br />";
    }

    return $itemlist;
}

1;
# End of Keyword.pm
